"""Create data head.

Revision ID: 59e1199aff7f
Revises:
Create Date: 2024-02-20T21:01:44.918799

"""

# revision identifiers, used by Alembic.
revision = "59e1199aff7f"
down_revision = None
branch_labels = ("data",)
depends_on = "048219045729"


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass
