{# Lazy workflows definition #}

LazyWorkflowInstance("workflows.{{ product.variable }}.create_{{ product.variable }}", "create_{{ product.variable }}")
LazyWorkflowInstance("workflows.{{ product.variable }}.modify_{{ product.variable }}", "modify_{{ product.variable }}")
LazyWorkflowInstance("workflows.{{ product.variable }}.terminate_{{ product.variable }}", "terminate_{{ product.variable }}")
LazyWorkflowInstance("workflows.{{ product.variable }}.validate_{{ product.variable }}", "validate_{{ product.variable }}")
