# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.domain.base import SubscriptionModel
from orchestrator.utils.docs import make_product_type_index_doc

SUBSCRIPTION_MODEL_REGISTRY: dict[str, type[SubscriptionModel]] = {}

__doc__ = make_product_type_index_doc(SUBSCRIPTION_MODEL_REGISTRY)

__all__ = ["SubscriptionModel", "SUBSCRIPTION_MODEL_REGISTRY"]
